package egs.wolves;

import abp.*;
import static util.Rand.pickElement;

import java.util.Random;

class Wolves {
  private static final float CATCH_REWARD = 1000f;
  private static final float MOVE_PENALTY = -1f;
  private static final int ITRS = 300000;

  public static void main(String[] args){
    AdaptiveProcess hunting = AdaptiveProcess.initialize();

    Adaptive<Grid,Move>
      w1 = hunting.newAdaptive(Move.SET),
      w2 = hunting.newAdaptive(Move.SET);
//      w1 = hunting.newAdaptive(Grid.ALL_POSSIBLE,Move.SET),
//      w2 = hunting.newAdaptive(Grid.ALL_POSSIBLE,Move.SET);
    Rabbit r = Rabbit.skittish();

    for (int i = 0; i < 10000; i++) {
      singleRound(hunting, w1, w2, r, CATCH_REWARD, MOVE_PENALTY, Grid.INITIAL);
    }


//    testScenerio(hunting, w1, w2, r,
//                  ITRS, 10000,
//                  CATCH_REWARD, MOVE_PENALTY, Grid.INITIAL);
//    exhaustiveLearn(hunting, w1, w2, Rabbit.skittish(),
//                      CATCH_REWARD, MOVE_PENALTY, 100);
    w1.setMode(Mode.GREEDY);
    w2.setMode(Mode.GREEDY);
    G_OFF = ITRS;
    testScenerio(hunting, w1, w2, r,
                  100000, 10000,
                  CATCH_REWARD, MOVE_PENALTY, Grid.INITIAL);

    while (true)
      verboseRound(hunting, w1, w2, Rabbit.interactive(),
                    CATCH_REWARD, MOVE_PENALTY, Grid.INITIAL);
  }

  static int G_OFF;

  private static void exhaustiveLearn(AdaptiveProcess hunting,
                                    Adaptive<Grid,Move> w1,
                                    Adaptive<Grid,Move> w2,
                                    Rabbit r,
                                    float catchReward,
                                    float movePenalty,
                                    int gamesPerState) {
    System.out.println("Learning from " + Grid.ALL.length * gamesPerState + " games");
    long s = System.currentTimeMillis();
    for (Grid w : Grid.ALL) {
      for (int ei = 0; ei < gamesPerState; ei++) {
        singleRound(hunting, w1, w2, r, catchReward, movePenalty, w);
      }
    }
    System.out.printf("  finished in %7.2f s%n", (System.currentTimeMillis() - s)/1000f);
  }

  // init == null -> we use a random initial value.
  private static void testScenerio(AdaptiveProcess hunting,
                                    Adaptive<Grid,Move> w1,
                                    Adaptive<Grid,Move> w2,
                                    Rabbit r,
                                    int numEpisodes,
                                    int sampleFreq,
                                    float reward,
                                    float penalty,
                                    Grid init) {
    double avgMoves = 0.0;
    double lastBlock = 0.0f;
    final Random rand = new Random(24);
    for (int ei = 1; ei <= numEpisodes; ei++) {
      int moves = singleRound(hunting, w1, w2, r, reward, penalty,
          init == null ? pickElement(rand,Grid.ALL) : init);

      avgMoves = (ei*avgMoves + moves) / (ei+1);
      lastBlock += moves;
      if (sampleFreq != 0 && ei % sampleFreq == 0) {
        System.out.printf("%8d %6.3f %6.3f %8d\n",
                      G_OFF + ei, avgMoves, lastBlock/sampleFreq, moves);
        lastBlock = 0.0;
      }
    }
  }

  private static int singleRound(AdaptiveProcess hunting,
          Adaptive<Grid,Move> w1,
          Adaptive<Grid,Move> w2,
          Rabbit r,
          float captureReward,
          float movePenalty,
          Grid w) {
    int numMoves = 0;
    while (true) {
      w = w.moveRabbit(r.pickMove(w));
      w = w.moveWolf1(w1.suggest(w));
      w = w.moveWolf2(w2.suggest(w));

      numMoves++;

      if (w.rabbitCaught())
        break;

      hunting.reward(movePenalty);
    }
    hunting.reward(captureReward);
    hunting.endEpisode();
    return numMoves;
  }

  private static int verboseRound(AdaptiveProcess hunting,
                                  Adaptive<Grid,Move> w1,
                                  Adaptive<Grid,Move> w2,
                                  Rabbit r,
                                  float reward,
                                  float penalty,
                                  Grid init) {
    Grid w = (init == null) ? pickElement(new Random(),Grid.ALL) : init;
    System.out.println(w);

    int mi = 0;
    while (true) {
      Move rm = r.pickMove(w);
      w = w.moveRabbit(rm);

      Move w1m = w1.suggest(w);
      w = w.moveWolf1(w1m);

      Move w2m = w2.suggest(w);
      w = w.moveWolf2(w2m);

      System.out.printf("MOVE %2d%n", mi);
      System.out.printf("r:  %8s%n", rm);
      System.out.printf("w1: %8s%n", w1m);
      System.out.printf("w2: %8s%n", w2m);
      System.out.println(w);

      if (w.rabbitCaught()) {
        System.out.println("RABBIT CAUGHT!");
        break;
      }

      hunting.reward(penalty);
      mi++;
    }
    hunting.reward(reward);
    hunting.endEpisode();

    return mi;
  }
}

